from .distributions.normalized_distribution import normalized_distribution
from .distributions.Distribution import Distribution
from .distributions.SumOfDiracs import SumOfDiracs
from .distributions.UnitBox import UnitBox

from .D2GTransportMap import D2GTransportMap
from .G2DTransportMap import G2DTransportMap
from .PowerDiagram import PowerDiagram

import numpy as np

class DisplayParameters:
    num_newton_iteration = 0
    nb_relaxation_steps = 1
    max_error_ratio = 2
    norm_2_error = 0
    max_error = 0
    max_dw = 0 # max delta weight

    def __init__( self, list = [ 'max_error_ratio' ] ):
        if list is not None:
            for n, attr in enumerate( list ):
                assert attr in self
                setattr( self, attr, n + 1 )

    def write( self, solver ): 
        l = [ 'max_error_ratio', 'max_error', 'norm_2_error', 'nb_relaxation_steps', 'max_dw' ]
        display_first_iteration = False
        displayed_something = False
        for attr in l:
            if getattr( self, attr ):
                if len( getattr( solver, attr + '_history' ) ):
                    value = getattr( solver, attr + '_history' )[ -1 ]
                    if isinstance( value, float ):
                        print( f'{ attr }: {value:.4e}', end=' ' )
                    else:
                        print( f'{ attr }: {value}', end=' ' )
                    displayed_something = True
                else:
                    display_first_iteration = True

        if displayed_something:
            if display_first_iteration:
                print( '(first iteration)', end=' ' )
            print()


class SdotSolver:
    """
        Class to find a transport plan between a sum of dirac and a density.

    """
    def __init__( self, source_measure, target_measure = None, norm = 2, display: DisplayParameters = None ):
        # internal attributes
        self._source_measure = None
        self._target_measure = None

        # type update
        if not isinstance( display, DisplayParameters ):
            display = DisplayParameters( display )

        # the PowerDiagram used for the computations stores the acceleration structure with the positions and the weights
        self.power_diagram = PowerDiagram()

        # stopping criteria
        self.max_error_ratio_target = None
        self.max_error_target = None

        # solver parameters
        self.max_nb_newton_iterations = 500
        self.max_nb_relaxation_steps = 20

        # error history
        self.nb_relaxation_steps_history = []
        self.max_error_ratio_history = []
        self.norm_2_error_history = []
        self.max_error_history = []
        self.max_dw_history = []

        self.raise_if_error = True
        self.display = display
        self.status = 'none'

        # properties and attributes from ctor arguments
        self.source_measure = source_measure
        self.target_measure = target_measure
        self.norm = norm

    def forward_map( self ):
        return D2GTransportMap()

    def backward_map( self ):
        return D2GTransportMap()

    def adjust_potentials( self ):
        return self.newton_solve()

    def newton_solve( self, relaxation_coefficient = 1.0 ):
        # first system
        m_rows, m_cols, m_vals, residual, error_code = self.newton_system()
        if error_code:
            return self._set_status( 'void cells before the first iteration' )

        # check convergence
        if self._convergence_test( residual, None ):
            return self._set_status( 'ok' )

        # newton iteration
        for _ in range( self.max_nb_newton_iterations ):
            # linear solve
            mw = self._system_solve( m_rows, m_cols, m_vals, residual )
            dw = relaxation_coefficient * mw

            # update the weights
            self.power_diagram._weights -= dw
            self.power_diagram._weight_have_been_modified() # TODO: keep the same acceleration structure

            # loop if void cell
            for nb_relax_divs in range( self.max_nb_relaxation_steps ):
                # solve again the linear system
                m_rows, m_cols, m_vals, residual, error_code = self.newton_system()

                if error_code == 0:
                    self.nb_relaxation_steps_history.append( nb_relax_divs )
                    break

                # smaller weights update
                dw /= 2
                self.power_diagram._weights += dw
                self.power_diagram._weight_have_been_modified() # TODO: keep the same acceleration structure
            else:
                return self._set_status( 'max_lg2_relaxation reached' )

            # check convergence
            if self._convergence_test( residual, mw ):
                break
        else:
            return self._set_status( 'max_nb_newton_iterations reached' )

        return self._set_status( 'ok' )
                

    def _convergence_test( self, residual, dw ):
        # update history
        self.max_error_ratio_history.append( np.max( residual ) * self.nb_unknowns )
        self.norm_2_error_history.append( np.linalg.norm( residual ) )
        self.max_error_history.append( np.max( residual ) )
        if dw is not None:
            self.max_dw_history.append( np.max( np.abs( dw ) ) )

        # display
        self.display.write( self )

        # test convergence
        ok = True
        tested = False
        if self.max_error_ratio_target is not None:
            ok = ok and self.max_error_ratio_history[ -1 ] <= self.max_error_ratio_target
            tested = True
        if self.max_error_target is not None:
            ok = ok and self.max_error_history[ -1 ] <= self.max_error_target
            tested = True

        if tested:
            return ok
        return self.max_error_ratio_history[ -1 ] <= 1e-4

    def _set_status( self, status ):
        self.status = status
        if self.raise_if_error and status != 'ok':
            raise SdotSolverError( status )
        return self.status

    # def measure_ratio_error( self, norm = 'inf' ):
    #     f = np.full( [ self.nb_unknowns ], 1 / self.nb_unknowns )
    #     m = self.power_diagram.measures()
    #     if norm == 'inf':
    #         return np.max( np.abs( m - f ) / f )
    #     if norm == 2:
    #         return np.linalg.norm( ( m - f ) / f )
    #     raise RuntimeError( "TODO" )

    # def n2_error( self ):
    #     f = np.full( [ self.nb_unknowns ], 1 / self.nb_unknowns )
    #     m = self.power_diagram.measures()
    #     return np.linalg.norm( m - f )

    def newton_system( self ):
        self._check_inputs()

        # data from the power diagram
        m_rows, m_cols, m_vals, v_vals, error_code = self.power_diagram.dintegrals_dweights()
        residual = v_vals - np.full( [ self.nb_unknowns ], 1 / self.nb_unknowns )

        # force a delta weight to be equal to 0 (the first one...)
        for i in range( m_rows.size ):
            if m_rows[ i ] == m_cols[ i ]:
                m_vals[ i ] *= 2
                break

        return m_rows, m_cols, m_vals, residual, error_code
    
    def _system_solve( self, m_rows, m_cols, m_vals, v_vals ):
        from scipy.sparse.linalg import spsolve
        from scipy.sparse import coo_matrix

        M = coo_matrix( ( m_vals, ( m_rows, m_cols ) ), shape = ( v_vals.size, v_vals.size ) ).tocsr()
        return spsolve( M, v_vals )

    def plot( self, plt = None ):
        if plt is None:
            import matplotlib.pyplot as plt
            self.plot( plt )
            plt.show()
            return

        self._check_inputs()
        self.power_diagram.plot( plt )

    @property
    def nb_unknowns( self ):
        self._check_inputs()
        return self.power_diagram._weights.size

    @property
    def target_measure( self ):
        return self._target_measure

    @target_measure.setter
    def target_measure( self, measure ):
        self._target_measure = self._normalize_and_use_measure( measure )

    @property
    def source_measure( self ):
        return self._source_measure

    @source_measure.setter
    def source_measure( self, measure ):
        self._source_measure = self._normalize_and_use_measure( measure )

    @property
    def ndim( self ):
        for dist in [ self._source_measure, self._target_measure ]:
            p = dist.ndim
            if p is not None:
                return p
        return None

    def _normalize_and_use_measure( self, distribution ):
        """ register distribution (use it as self.power_diagram.positions or self.power_diagram.underlying_measure) """
        if distribution is None:
            return distribution

        # check type
        if not isinstance( distribution, Distribution ):
            # for this class, we consider that as np.ndarray represents a sum of uniformly weighted diracs
            if isinstance( distribution, np.ndarray ):
                distribution = SumOfDiracs( distribution )
            else:
                distribution = normalized_distribution( distribution )

        # register power diagram data        
        if isinstance( distribution, SumOfDiracs ):
            self.power_diagram.positions = distribution.positions
        else:
            self.power_diagram.underlying_measure = distribution

        # return the normalized version
        return distribution


    def _check_inputs( self ):
        """ return ( generic_dist, dirac_dist, dirac_is_target ) """

        if self._target_measure is None:
            self.target_measure = UnitBox()

        # check distribution compatibility
        sd = isinstance( self._source_measure, SumOfDiracs )
        td = isinstance( self._target_measure, SumOfDiracs )
        if sd + td == 0:
            raise RuntimeError( "SdotSolver expects at least one distribution to be a SumOfDiracs" )
        if sd + td == 2:
            raise RuntimeError( "SdotSolver expects at most one distribution to be a SumOfDiracs" )

        # distributions in the expected ordering
        if td:
            generic_dist, dirac_dist, dirac_is_tgt = self._source_measure, self._target_measure, True
        else:
            generic_dist, dirac_dist, dirac_is_tgt = self._target_measure, self._source_measure, False

        # output
        return generic_dist, dirac_dist, dirac_is_tgt
    

class SdotSolverError( RuntimeError ):
    pass



from .distributions.SymbolicFunction import SymbolicFunction
from .distributions.Distribution import Distribution
from .distributions.SumOfDiracs import SumOfDiracs
from .distributions.UnitBox import UnitBox

from .bindings.loader import set_auto_rebuild, global_verbosity_level

from .optimal_transport_plan import optimal_transport_plan

from .D2GTransportMap import D2GTransportMap
from .G2DTransportMap import G2DTransportMap
from .SdotSolver import SdotSolver

from .PowerDiagram import PowerDiagram
from .VtkOutput import VtkOutput
from .PoomVec import PoomVec
from .Expr import Expr
from .Cell import Cell


import numpy as np

class TransformationMatrix:
    def __init__( self, value = None ):
        if isinstance( value, TransformationMatrix ):
            self.value = value.value
        else:
            self.value = value

    def get( self, ndim = None ):
        # None
        if self.value is None:
            return np.eye( ndim + 1 )

        # ( M, V )
        if isinstance( self.value, tuple ):
            assert( len( self.value ) == 2 )
            M = np.array( self.value[ 0 ] )
            V = np.array( self.value[ 1 ] )
            
            assert( M.shape[ 0 ] == M.shape[ 1 ] )
            assert( M.shape[ 0 ] == V.shape[ 0 ] )
            s = M.shape[ 0 ]

            res = np.eye( s + 1 )
            res[ :s, :s ] = M
            res[ :s, s ] = V
            return res
        
        # full matrix
        array = np.array( self.value )
        if len( array.shape ) == 2:
            assert( array.shape[ 0 ] == array.shape[ 1 ] )
            if ndim is not None:
                assert( array.shape[ 0 ] == ndim + 1 )
            return array
        
        # vector
        if len( array.shape ) == 1:
            s = array.size

            res = np.eye( s + 1 )
            res[ :s, s ] = array
            return res
        
        raise ValueError( "invalid transformation" )



from .SymbolicFunction import SymbolicFunction
from .Distribution import Distribution
from .UnitBox import UnitBox

def normalized_distribution( dist, return_unit_box_if_not_defined = True ):
    """ return a Distribution instance from `dist`, which can be a scalar, a string, ... """

    # default value
    if dist is None:
        if return_unit_box_if_not_defined:
            return UnitBox()
        return None
    
    # already the good type
    if isinstance( dist, Distribution ):
        return dist 

    # scalar    
    if isinstance( dist, float ) or isinstance( dist, int ) or isinstance( dist, str ):
        return SymbolicFunction( dist )

    # not found    
    raise ValueError( f"don't know how to make a distribution from a { type( dist ) }" )



# from .Cell import Cell
import numpy as np

class PowerDiagramSummary:
    """
        Aggregation of arrays that represents a fully computed PowerDiagram.

        In these arrays, if < `nb_seeds`, a `ref` represents a cell index. Else (if >= `nb_seeds`) it represents a boundary index + `nb_seeds` (i.e. `boundary_index = ref - summary.nb_seeds`).
        
        Attributes:
            `vertex_coords`: coordinates for each vertex
            `ref_lists`: `ref_lists[ d ]` gives the list of refs for each item of dimensionality `d`. For instance, `ref_lists[ 0 ]` gives the refs (cells of boundary indices) for each vertex, `ref_lists[ 1 ]` gives the refs for each edge, ... It describes how items has been made (from which cell and boundary intersections)
            `parenting`: `parenting[ r ][ c ]` gives parenting (parents, children or siblings depending or `r - c`) of items of dimensionality `c` for  items of dimensionality `r`
            `base`: can be use if `true_dimensionality` < `ndim` to get coordinate in the `ndim` space

    """
    def __init__( self,vertex_coords, ref_lists, parenting, boundary_items, base ):
        self.boundary_items = boundary_items
        self.vertex_coords  = vertex_coords
        self.ref_lists      = ref_lists
        self.parenting      = parenting
        self.base           = base

    @property
    def nb_seeds( self ):
        return len( self.vertex_coords )

    @property
    def true_dimensionality( self ):
        return self.base.shape[ 1 ]

    @property
    def ndim( self ):
        return self.base.shape[ 0 ]

    def nb_items( self, dim ):
        """ nb items of dimensionality `dim` """
        return len( self.parenting[ dim ][ 0 ] )

    def barycenters( self, dim = None ):
        """ by default, return the barycenters of the cells.

            specifiying dim allows for returning the barycenters of vertices (dim=0), edges (dim=1), and so on.

            TODO: underlying_measures
        """
        if dim is None:
            dim = self.ndim

        if dim == 0:
            return self.vertex_coords

        res = np.empty( [ self.nb_items( dim ), self.ndim ] )
        for n, vs in enumerate( self.parenting[ dim ][ 0 ] ):
            res[ n, : ] = np.mean( self.vertex_coords[ vs ], axis=0 )
 
        return res
 
    def integrals( self ):
        """  """
        raise RuntimeError( "TODO" )
    
    


from urllib.parse import quote_plus
from pathlib import Path
from munch import Munch
import archspec.cpu
import subprocess
import platform
import numpy
import sys
import os

# needed to share the symbols
sys.setdlopenflags( os.RTLD_GLOBAL | os.RTLD_LAZY )

# ------------------------------------------------------------------------------------------------------------------------------------------
def get_global_directory_for( name ):
    """
      find a directory to 
        * copy .cpp/.h
        * install ext libraries
        * store .so/.dylib/... files
    """
    
    # Try in source/build
    sources = Path( __file__ ).parent.parent
    res = sources / name
    if can_use_dir( res ):
        return res
    
    # Else, try in $HOME/sdot_build
    local = Path.home() / ".local" / "lib" / "sdot"
    res = local / name
    if can_use_dir( res ):
        return res    
    
    raise RuntimeError( f"found no way to make a { name } dir (tried in { sources } and in { local }" )

def can_use_dir( dir ):
    try:
        os.makedirs( dir, exist_ok = True )
        return os.access( dir, os.W_OK )
    except OSError:
        pass
    return False

# ------------------------------------------------------------------------------------------------------------------------------------------
def push_activity_log( name ):
    activities.append( name )
    print( ' -> '.join( activities ), end = '' )
def pop_activity_log():
    s = ' -> '.join( activities )
    print( '\r' + ' ' * len( s ), end = '\r' )
    activities.pop()


def set_auto_rebuild( value ):
    global auto_rebuild
    auto_rebuild = value

def normalized_dtype( dtype, raise_exception_if_not_found = True ):
    if dtype in [ "FP64", "double", numpy.float64, numpy.int32, numpy.int64 ]:
        return "FP64"
    if dtype in [ "FP128", numpy.float128 ]:
        return "FP128"
    if dtype in [ "FP32", "float", numpy.float32 ]:
        return "FP32"
    if raise_exception_if_not_found:
        raise RuntimeError( f"unknown dtype `{ dtype }`" )
    return None


def numpy_dtype_for( dtype ):
    if dtype in [ "FP128" ]:
        return numpy.float128
    if dtype in [ "FP64" ]:
        return numpy.float64
    if dtype in [ "FP32" ]:
        return numpy.float32
    return None


def type_score( dtype: str ):
    if dtype.startswith( "FP" ):
        return int( dtype[ 2: ] )
    raise RuntimeError( "TODO" )


def type_promote( dtypes ):
    if len( dtypes ) == 0:
        return None

    res = normalized_dtype( dtypes[ 0 ] )
    for dtype in map( normalized_dtype, dtypes[ 1: ] ):
        if type_score( dtype ) > type_score( res ):
            res = dtype

    return res


def get_local_build_directory( global_build_directory, *kargs ):
    res = global_build_directory
    for arg in kargs:
        res = res / arg

    os.makedirs( res, exist_ok = True )
    if os.access( res, os.W_OK ):
        return res
    
    raise RuntimeError( f"failed to make the local build dir '{ res }'" )


def module_for( name, dir_of_the_SConstruct_file = Path( __file__ ).parent, use_arch = False, **kwargs ):
    # ensure that the std objects are loaded
    if name != 'generic_objects':
        module_for( 'generic_objects', dir_of_the_SConstruct_file, False )

    # sorted list of parameters
    plist = [ ( x, str( y ) ) for x, y in kwargs.items() ]
    plist.sort()

    # append 'os'
    plist.append( ( 'os', platform.system() ) )
    # if use_arch:
    #     plist.append( ( 'arch', archspec.cpu.generic_microarchitecture() ) )

    # other presentation for the parameters
    ilist = [ f"{ k }={ quote_plus( str( v ) ) }" for k, v in plist ]
    vlist = [ quote_plus( str( v ) ) for _, v in plist ]

    # names
    suffix = "_".join( vlist ).replace( "%", '_' )
    module_name = f'{ name }_bindings_for_{ suffix }'

    # in the cache
    if module_name in loader_cache:
        return loader_cache[ module_name ]

    # where to find/put the files
    local_build_directory = get_local_build_directory( global_build_directory, name, suffix )
    if local_build_directory not in sys.path:
        sys.path.insert( 0, str( local_build_directory ) )

    # if allowed, try to load directly the library
    module = None
    if not auto_rebuild:
        try:
            module = __import__( module_name )
        except ModuleNotFoundError:
            pass

    # else, build and load it
    if module is None:
        if global_verbosity_level:
            push_activity_log( f"compilation of { name } for { kwargs }" )

        #
        source_directory = Path( __file__ ).parent

        args = [ 'scons', '-j8', f"--sconstruct={ dir_of_the_SConstruct_file / ( name + '.SConstruct' ) }", # '-s',
            f"source_directory={ source_directory }", 
            f"ext_directory={ global_ext_directory }", 
            f"module_name={ module_name }", 
            f"suffix={ suffix }",
        ] + ilist

        print( " ".join( args ) )

        # call scons
        ret_code = subprocess.call( args, cwd = str( local_build_directory ), shell = False )
        
        if ret_code:
            raise RuntimeError( f"Failed to compile the C++ { name } binding" )

        if global_verbosity_level:
            pop_activity_log()

        # import
        print('impot', module_name )
        module = __import__( module_name )

    # change names
    res = Munch()
    for n, v in module.__dict__.items():
        if n.endswith( '_' + suffix ):
            n = n[ : -1 - len( suffix ) ]
        res[ n ] = v

    # store the module in the cache
    loader_cache[ module_name ] = res

    return res

# --------------------------------------- global variable ---------------------------------------
global_build_directory = get_global_directory_for( "build" )
global_ext_directory = get_global_directory_for( "ext" )
global_verbosity_level = 1
auto_rebuild = False
loader_cache = {}
activities = []



from .bindings.loader import normalized_dtype, numpy_dtype_for, type_promote, module_for
from .distributions.normalized_distribution import normalized_distribution
from .bindings.integration_module import integration_module
from .TransformationMatrix import TransformationMatrix
from .PowerDiagramSummary import PowerDiagramSummary
from .VtkOutput import VtkOutput
from .PoomVec import PoomVec
from .Expr import Expr
from .Cell import Cell

from munch import Munch
import numpy as np



class PowerDiagram:
    """
        Stores all the data needed to compute a power diagram, i.e.
            * seed positions
            * seed weights
            * optional boundaries
            * an optional underlying measure and an optional underlying radius function for method like integration, barycenter, ...

        Boundaries can be represented as
            * an array [ [ dir_{num_axis}, ..., val ]_{num_boundary}, ... ] where a point `x` is considered as exterior if `np.dot( dir, point ) > val`.
            * a tuple of arrays, one for the directions (each row has a direction), and one for the values (a 1D array)

        Underlying measure is internally stored as an instance of `Distribution`. It is used to compute measures (volumes/areas/...) and can give boundaries (for instance, IndicatorFunction(UnitBox()) will naturally add boundaries).

        `dtype` is the name of the scalar type used by the bindings. One can use numpy types as input (e.g. numpy.float64) or string representation (e.g. "FP32").

        By default, the scalar type and the dimension used by the bindings are deduced from the inputs (positions, weights, boundaries, ...), 
        but is can be forced by specifying `dtype` and `ndim` in the ctor, or by setting them using `self.dtype = ...` (dtype is a property) or `self.ndim = ...`

        Internally, positions and weights are stored in PoomVecs (Potientially Out of Memory vectors). They can be defined for instance using numpy arrays.

        If `automatic_conversion_to_ndarrays` is set to True, output arrays are automically converted to numpy `ndarray`.
        For instance, `self.positions` will return a `ndarray` instead of a `PoomVec`

    """
    def __init__( self, positions = None, weights = None, boundaries = None, underlying_measure = None, dtype = None, ndim = None, automatic_conversion_to_ndarrays = True ):
        # base init
        self._density_under_final_boundaries = None # symbolic expression representing the value inside the convex boundaries 
        self._periodicity_transformations = [] # list of TransformationMatrix 
        self._acceleration_structure = None # paving + hierarchical repr of positions and weights
        self._underlying_measure = None # instance of Distribution
        self._binding_module = None
        self._boundaries = None # expected to be a ndarray
        self._positions = None # PoomVec
        self._weights = None # PoomVec
        self._dtype = None # can be used to force the data type
        self._ndim = None # can be used to force the dimension

        # prefs
        self.automatic_conversion_to_ndarrays = automatic_conversion_to_ndarrays

        # ctor arguments phase 1
        self.dtype = dtype # used to force the type
        self.ndim = ndim # used to force the number of dimensions

        # ctor arguments phase 2
        self.underlying_measure = underlying_measure
        self.boundaries = boundaries
        self.positions = positions
        self.weights = weights

    def add_box_boundaries( self, min_offset = None, max_offset = None, ndim = None, base = None ):
        """ delimit the power diagram by an (hyper-)cube (a square in 2D, and so on)

            if ndim is not specified, one tries to guess it from the defined attributes

            min_offset and max_offset are related to base. They can be single scalars or vector, with one value for each base item.

            If base is not specified, one takes the identity matrix
        """

        # check argument values
        if base is not None:
            ndim = len( base )

        if ndim is None:
            ndim = self.ndim
            if ndim is None:
                raise RuntimeError( "Found no way to guess ndim" )
            
        if base is None:
            base = np.identity( ndim )

        if min_offset is None:
            min_offset = 0
        if max_offset is None:
            max_offset = 1
        if isinstance( min_offset, ( int, float ) ):
            min_offset = np.full( [ ndim ], min_offset )
        if isinstance( max_offset, ( int, float ) ):
            max_offset = np.full( [ ndim ], max_offset )

        # update the boundaries
        l = []
        if self._boundaries is not None:
            l = list( self._boundaries )
        for n in range( ndim ):
            l.append( [ -( d == n ) for d in range( ndim ) ] + [ - min_offset[ n ] ] )
            l.append( [ +( d == n ) for d in range( ndim ) ] + [ + max_offset[ n ] ] )
        self.boundaries = l

    @property
    def periodicity_transformations( self ):
        return self._periodicity_transformations
    
    @periodicity_transformations.setter
    def periodicity_transformations( self, values ):
        self._periodicity_transformations = [ TransformationMatrix( v ) for v in values ]
        self._acceleration_structure = None

    @property
    def boundaries( self ):
        return self._boundaries

    @boundaries.setter
    def boundaries( self, values ):
        if values == None:
            return
        
        if type( values ) is tuple:
            dirs = np.asarray( values[ 0 ] )
            vals = np.asarray( values[ 1 ] )

            assert dirs.ndim == 2
            assert vals.ndim == 1
            assert dirs.shape[ 0 ] == vals.shape[ 0 ]

            self._boundaries = np.ascontiguousarray( np.concatenate( ( dirs, vals.reshape( [ -1, 1 ] ) ), axis = 1 ) )
        else:
            self._boundaries = np.ascontiguousarray( values )

    @property
    def positions( self ):
        if self.automatic_conversion_to_ndarrays and self._positions is not None:
            return self._positions.as_ndarray
        return self._positions

    @positions.setter
    def positions( self, values ):
        if values is None:
            return
        self._acceleration_structure = None
        self._positions = PoomVec( values )

    @property
    def weights( self ):
        if self.automatic_conversion_to_ndarrays and self._weights is not None:
            return self._weights.as_ndarray
        return self._weights

    @weights.setter
    def weights( self, values ):
        if values is None:
            return
        self._weights = PoomVec( values )
        self._weight_have_been_modified()

    def _weight_have_been_modified( self ):
        self._acceleration_structure = None # TODO: update weight without redoing everything

    @property
    def underlying_measure( self ):
        return self._underlying_measure

    @underlying_measure.setter
    def underlying_measure( self, values ):
        self._underlying_measure = normalized_distribution( values )

    @property
    def dtype( self ):
        # if specified by the user
        if self._dtype:
            return self._dtype
        
        # else, use the dtypes from the inputs
        dtypes = []
        for v in [ self._boundaries, self._positions, self._weights ]:
            if v is not None:
                dtypes.append( normalized_dtype( v.dtype ) )

        return type_promote( dtypes )

    @dtype.setter
    def dtype( self, value ):
        if value == None:
            return
        self._dtype = normalized_dtype( value )

    @property
    def ndim( self ):
        # if specified by the user
        if self._ndim:
            return self._ndim
        
        # else, use the ndim from the inputs
        if self._boundaries is not None and len( self._boundaries ):
            return self._boundaries.shape[ 1 ] - 1
        if self._positions is not None:
            return self._positions.shape[ 1 ]

        return None

    @ndim.setter
    def ndim( self, value ):
        if value == None:
            return
        self._ndim = int( value )

    def for_each_cell( self, function, max_nb_threads = None ):
        """
            Call `function( cell: Cell )` for each cell in the power diagram.

            If max_nb_threads == 1, `function` will be called from the main thread (which can be useful for instance for matplotlib, etc...)
        """

        if not self._update_internal_attributes():
            return

        # call the compiled function and wrap the cell
        cwc = lambda cell: function( Cell( _cell = cell, _binding_module = self._binding_module ) )
        self._acceleration_structure.for_each_cell( self._base_cell._cell, cwc, max_nb_threads or 0 )

    def cell_integrals( self, func = 1 ):
        """ get integration of `func` other each cell """
        if not self._update_internal_attributes():
            return np.empty( [ 0, self.ndim or 0 ] )

        # final expression to integrate other the cell
        final_expr = self._density_under_final_boundaries * Expr( func )

        cv = final_expr.constant_value()
        if cv is not None:
            return self._binding_module.measures( self._acceleration_structure, self._base_cell._cell, cv )

        module, rt_data = integration_module( final_expr, self.dtype, self.ndim )
        return module.power_diagram_integrals( self._acceleration_structure, self._base_cell._cell, rt_data )

    def dintegrals_dweights( self, func = 1 ):
        """ 
            get derivatives of integration of `func` other each cell

            return ( m_rows, m_cols, m_vals, v_vals, error_code ) with m = matrix, v = vector
        """
        if not self._update_internal_attributes():
            raise ( [], [], [], [], 0 )
        
        # final expression to integrate other the cell
        final_expr = self._density_under_final_boundaries * Expr( func )

        cv = final_expr.constant_value()
        if cv is not None:
            return self._binding_module.dmeasures_dweights( self._acceleration_structure, self._base_cell._cell, cv )

        raise RuntimeError( "TODO" )
    
    def summary( self ):
        """ return a PowerDiagramSummary with arrays that contain global information to summarize the power diagram (coords, parenting, ...).
        """
        if not self._update_internal_attributes():
            raise ValueError( "missing some attribute to be able to make a summary" )

        # TODO: PowerDiagram in sub-spaces
        s = self._binding_module.summary( self._acceleration_structure, self._base_cell._cell )
        return PowerDiagramSummary( *s, np.eye( self.ndim ) )

    def plot_pyplot( self, fig, **kwargs ):
        """  
            For optional arguments, see the `Cell.plot` method
        """
        if fig is None:
            import matplotlib.pyplot as plt
            self.plot( plt, **kwargs )
            plt.show()
            return

        self.for_each_cell( lambda cell: cell.plot( fig, **kwargs ), max_nb_threads = 1 )
        return fig

    def plot_vtk( self, vtk_output, **kwargs ):
        if isinstance( vtk_output, str ):
            vo = VtkOutput()
            self.plot_vtk( vo, **kwargs )
            vo.save( vtk_output )
            return 
        
        assert isinstance( vtk_output, VtkOutput )

        self._update_internal_attributes()
        self._binding_module.plot_vtk( vtk_output._inst, self._acceleration_structure, self._base_cell._cell )

    def plot( self, fig = None, **kwargs ):
        """  
            plot cells content in
            * pyplot figure (`pyplot` or in a `pyplot.fig` for instance)
            * or in a VtkOutput

            For optional arguments, see the `Cell.plot` method
        """
        # if fig is None:
        #     import matplotlib.pyplot as plt
        #     self.plot( plt )
        #     plt.show()
        #     return

        if isinstance( fig, str ):
            if fig.endswith( ".vtk" ):
                return self.plot_vtk( fig, **kwargs )
            raise ValueError( "Unhandled file type" )

        return self.plot_pyplot( fig, **kwargs )

    def _update_internal_attributes( self ):
        """ 
            compute self._binding_module, self._acceleration_structure and self._base_cell if not already done
        """

        # binding module
        dtype = self.dtype
        ndim = self.ndim
        if dtype is None or ndim is None:
            return False
        self._binding_module = module_for( 'sdot', use_arch = True, scalar_type = dtype, nb_dims = ndim )

        # acceleration structure
        if self._acceleration_structure is None or self._acceleration_structure.dtype != dtype or self._acceleration_structure.ndim != ndim:
            # set weights if not defined
            if self._weights is None:
                self._weights = PoomVec( np.zeros( self._positions.shape[ 0 ], dtype = numpy_dtype_for( dtype ) ) )

            # get dtype consistency. TODO: use PoomVecInst_Cast<...>
            if self._positions.dtype != dtype:
                self._positions = PoomVec( self._positions.as_ndarray.astype( numpy_dtype_for( dtype ) ) )
            if self._weights.dtype != dtype:
                self._weights = PoomVec( self._weights.as_ndarray.astype( numpy_dtype_for( dtype ) ) )

            #
            transformation_matrices = []
            for i, pt in enumerate( self._periodicity_transformations ):
                tm = pt.get( ndim )
                ti = np.linalg.inv( tm )
                transformation_matrices.append( tm )
                transformation_matrices.append( ti )
                for j in range( i ):
                    om = self._periodicity_transformations[ j ].get( ndim )
                    oi = np.linalg.inv( om )
                    transformation_matrices.append( om @ tm )
                    transformation_matrices.append( oi @ tm )
                    transformation_matrices.append( om @ ti )
                    transformation_matrices.append( oi @ ti )

            # make self._acceleration_structure
            self._acceleration_structure = self._binding_module.LowCountAccelerationStructure( self._positions._vec, self._weights._vec, transformation_matrices )

        # beginning of base cell
        self._base_cell = Cell( _cell = self._binding_module.Cell(), _binding_module = self._binding_module )
        nb_base_bnds = 0
        if self._boundaries is not None:
            ndim = self._binding_module.ndim()
            nb_base_bnds = len( self._boundaries )
            for n, bnd in enumerate( self._boundaries ):
                self._base_cell.cut_boundary( bnd[ : ndim ],  bnd[ ndim ], n )

        # density split between boundaries and value (possible modification of base cell)
        bnds, value = self._underlying_measure.boundary_split( ndim )
        self._density_under_final_boundaries = Expr( value )

        for n, bnd in enumerate( bnds ):
            self._base_cell.cut_boundary( bnd[ : ndim ],  bnd[ ndim ], nb_base_bnds + n )

        return True




from urllib.parse import unquote_plus
from pathlib import Path
from munch import Munch
import subprocess
import sysconfig
import pybind11
import sys
import os

# helper to get n-th parent directory
def pdir( dir, n = 1 ):
    if n == 0:
        return dir
    return pdir( os.path.dirname( dir ), n - 1 )

def args_to_obj( ARGLIST ):
    args = Munch()
    for k, v in ARGLIST:
        if k in [ 'module_name', 'suffix' ]:
            args[ k ] = v
        else:
            args[ k ] = unquote_plus( v )
    return args

def download_and_unzip( link, src, dst, ext_directory ):
    import dload
 
    print( f"Downloading { dst } in { ext_directory }" )

    dload.save_unzip( link, str( ext_directory ), delete_after = True )
    try:
        os.rename( ext_directory / src, ext_directory / dst )
    except OSError:
        pass

# def git_clone( link ):
#     from git import Repo  # pip install gitpython

#     Repo.clone_from( ,  )

def construct( Environment, VariantDir, Configure, ARGLIST, name, used_arg_names, files ):
    # common args
    args = {}
    for k, v in ARGLIST:
        if k not in [ 'ext_directory', 'source_directory', 'module_name', 'suffix' ]:
            args[ k ] = unquote_plus( v )
        else:
            args[ k ] = v

    source_directory = Path( args[ 'source_directory' ] )
    ext_directory = Path( args[ 'ext_directory' ] )
    module_name = args[ 'module_name' ]
    suffix = args[ 'suffix' ]

    # build directory
    VariantDir( 'build', source_directory, duplicate=False )
    VariantDir( 'ext', ext_directory, duplicate=False )

    # includes
    CPPPATH = [
        # src
        os.path.join( source_directory, 'cpp' ),

        # ext
        os.path.join( ext_directory, 'tl20', 'src', 'cpp' ),
        os.path.join( ext_directory, 'asimd', 'src' ),
        os.path.join( ext_directory, 'boost' ),
        os.path.join( ext_directory, 'eigen' ),

        # system
        sysconfig.get_paths()[ 'include' ], # Python.h
        pybind11.get_include(), # pybind11.h
    ]


    # CXXFLAGS
    CXXFLAGS = [
        f'-DSDOT_CONFIG_module_name={ module_name }',
        f'-DSDOT_CONFIG_suffix={ suffix }',

        '-Wdeprecated-declarations',
        '-std=c++20',
        '-fopenmp',

        '-fdiagnostics-color=always',
        
        '-ffast-math',
        '-O3',

        '-g3',
    ]

    if 'arch' in used_arg_names:
       CXXFLAGS.append( '-march=' + args[ 'arch' ].replace( '_', '-' ) )

    for name in used_arg_names:
        CXXFLAGS.append( f'"-DSDOT_CONFIG_{ name }={ args[ name ] }"' )

    # LIBS
    LIBS = [
        # 'Catch2Main',
        # 'Catch2',
        # 'gomp',
    ]

    # LIBPATH
    LIBPATH = [
        # '/home/hugo.leclerc/.vfs_build/ext/Catch2/install/lib',
        # '/home/leclerc/.vfs_build/ext/Catch2/install/lib'
    ]

    # .cpp files
    sources = files + [
    ]

    # Environment
    env = Environment( CPPPATH = CPPPATH, CXXFLAGS = CXXFLAGS, LIBS = LIBS, LIBPATH = LIBPATH, SHLIBPREFIX = '' )

    # check the libraries
    conf = Configure( env )
    if not conf.CheckCXXHeader( 'boost/multiprecision/cpp_int.hpp' ):
        download_and_unzip( "https://archives.boost.io/release/1.86.0/source/boost_1_86_0.zip", "boost_1_86_0", "boost", ext_directory )
    if not conf.CheckCXXHeader( 'Eigen/Dense' ):
        download_and_unzip( "https://gitlab.com/libeigen/eigen/-/archive/3.4.0/eigen-3.4.0.zip", "eigen-3.4.0", "eigen", ext_directory )
    if not conf.CheckCXXHeader( 'asimd/SimdVec.h' ):
        download_and_unzip( "https://github.com/hleclerc/asimd/archive/refs/tags/asimd-v0.0.1-alpha.zip", "asimd-asimd-v0.0.1-alpha", "asimd", ext_directory )
    if not conf.CheckCXXHeader( 'tl/support/Displayer.h' ):
        download_and_unzip( "https://github.com/hleclerc/tl20/archive/refs/tags/v0.0.2.zip", "tl20-0.0.2", "tl20", ext_directory )
    env = conf.Finish()

    # register the library
    env.SharedLibrary( module_name + env[ 'SHLIBSUFFIX' ], list( map( str, sources ) ) )





class Distribution:
    """
        Ancestor class for distributions


    """
    
    def boundary_split( self, ndim = None ):
        """ get a tuple with
              * a list of convex boundaries (can be a void list)
              * the value inside the boundaries (something that can be converted to an Expr, but without any trivial convex boundaries)
        """
        return None
    
    @property
    def ndim( self ):
        """ return the prescribed dimensionality if there's any """
        return None 
    



from .bindings.loader import module_for, normalized_dtype, type_promote
from .TransformationMatrix import TransformationMatrix
# from types import ModuleType
import numpy as np

class Expr:
    """ wrapper around cpp sdot::Expr class to store symbolic expressions """

    def __init__( self, value = None ):
        """ 
        """
        # default value
        if value is None:
            value = 0

        # already an Expr ?
        if isinstance( value, Expr ):
            self._expr = value._expr
            return

        _module = module_for( 'generic_objects' )
        if isinstance( value, _module.Expr ):
            self._expr = value
            return

        # else, call the cpp ctor
        self._expr = _module.Expr( value )

    def subs( self, symbol_map ):
        map = {}
        for k, v in symbol_map.items():
            map[ k ] = Expr( v )._expr
        return Expr( self._expr.subs( map ) )

    def __getitem__( self, args ):
        """ assumes args are space variables """
        m = {}
        if isinstance( args, tuple ):
            for i, arg in enumerate( args ):
                m[ f'x_{ i }' ] = arg
        else:
            m[ 'x_0' ] = args
        return self.subs( m )

    # def __eq__( self, other ):
    #     if not isinstance( other, Expr ):
    #         return self.__eq__( Expr( other ) )
    #     return self._expr == other._expr

    def always_equal( self, that ):
        if not isinstance( that, Expr ):
            return self.always_equal( Expr( that ) )
        return self._expr.always_equal( that._expr )

    def constant_value( self ):
        """ if not constant, return None. Else, return the value """
        valid, value = self._expr.constant_value()
        if valid:
            return value
        return None

    @staticmethod
    def img_interpolation( array, transformation_matrix = None, interpolation_order = 0 ):
        """ symbolic expression from image

            Beware: it follows the numpy convention for the axes (for 2D, x_0 is the y axis, x_1 is the x axis)  
        """
        # need to load the module to get the Expr type
        module_for( 'generic_objects' )
        
        array = np.ascontiguousarray( array )
        trinv = np.linalg.inv( TransformationMatrix( transformation_matrix ).get( array.ndim ) )
        module = module_for( 'img_interpolation', 
            scalar_type = type_promote([ array.dtype, trinv.dtype ]), 
            interpolation_order = interpolation_order,
            nb_dims = array.ndim 
        )
        return Expr( module.Expr_from_image( array, trinv ) )

    @staticmethod
    def list_from_compact_repr( crepr ):
        _module = module_for( 'generic_objects' )
        return _module.expr_list_from_compact_repr( crepr )
    
    @staticmethod
    def as_expr( expr ):
        if expr is None:
            return Expr( "0" )
        if isinstance( expr, Expr ):
            return expr
        if isinstance( expr, list ):
            return [ Expr.as_expr( v ) for v in expr ]
        return Expr( expr )

    @staticmethod
    def ct_rt_split_of_list( expr_list ):
        module = module_for( 'generic_objects' )
        return module.ct_rt_split_of_list( [ Expr.as_expr( e )._expr for e in expr_list ] )

    @staticmethod
    def cell_splits_of_list( funcs, rt_data ):
        """ splits, final_funcs """
        module = module_for( 'generic_objects' )
        return module.cell_splits_of_list( [ Expr.as_expr( e )._expr for e in funcs ], rt_data )

    def ct_repr( self ):
        return self._expr.ct_repr()

    def rt_data( self ):
        return self._expr.rt_data()

    def boundary_split( self, ndim ):
        return self._expr.boundary_split( ndim )

    def __repr__( self ):
        return self._expr.__repr__()

    def __add__( self, that ):
        if not isinstance( that, Expr ):
            that = Expr( that )
        return Expr( self._expr.add( that._expr ) )

    def __sub__( self, that ):
        if not isinstance( that, Expr ):
            that = Expr( that )
        return Expr( self._expr.sub( that._expr ) )

    def __mul__( self, that ):
        if not isinstance( that, Expr ):
            that = Expr( that )
        return Expr( self._expr.mul( that._expr ) )

    def __truediv__( self, that ):
        if not isinstance( that, Expr ):
            that = Expr( that )
        return Expr( self._expr.div( that._expr ) )
    
    def __pow__( self, that ):
        if not isinstance( that, Expr ):
            that = Expr( that )
        return Expr( self._expr.pow( that._expr ) )




from .bindings.integration_module import integration_module
from .bindings.loader import module_for
from .Expr import Expr

from types import ModuleType
import numpy as np

class Cell:
    """ wrapper around cpp sdot::Cell class """

    def __init__( self, ndim = None, dtype = None, _cell = None, _binding_module = None ):
        """ _cell is the sdot::Cell class instance. If not specified, __init__ makes a new one using `ndim` and `dtype`.
            By default, dtype = np.double, ndim = 2
        """

        if _cell:
            assert _binding_module is not None
            self._binding_module = _binding_module
            self._cell = _cell
            return

        if ndim is None:
            ndim = 2

        if dtype is None:
            dtype = "FP64"

        self._binding_module = module_for( 'sdot', use_arch = True, scalar_type = dtype, nb_dims = ndim )
        self._cell = self._binding_module.Cell()

    def __repr__( self ):
        return self._cell.__repr__()

    def integral( self, funcs = '1', underlying_measure_override = None, underlying_radial_function_override = None ):
        """ integral of a symbolic function or a list of symbolic functions """
        if not isinstance( funcs, list ):
            return self.integral( [ funcs ], underlying_measure_override, underlying_radial_function_override )[ 0 ]
        
        # module généré
        module, rt_data = integration_module( funcs, self._binding_module.dtype(), self._binding_module.ndim() )
        return module.cell_integral( self._cell, rt_data )

    @property
    def true_dimensionality( self ):
        """ Size of the base.
        
            at the beginning, all the cell starts with true_dimensionality == 0. A first cut will lead to true_dimensionality == 1. 
            If not in the same direction, true_dimensionality will be incremented, ...
            If true_dimensionality < ndims, coordinates are expressed in self.base
        """
        return self._cell.true_dimensionality()

    @property
    def vertex_coords_td( self ):
        """ get list of coordinates for each vertex (stored in a numpy array) in the current base """
        return self._cell.vertex_coords( True )

    @property
    def vertex_coords( self ):
        """ get list of coordinates for each vertex (stored in a numpy array) """
        return self._cell.vertex_coords( False )

    @property
    def vertex_refs_td( self ):
        """ get list of cut indices for each vertex (stored in a numpy array) in the current base """
        return self._cell.vertex_refs( True )

    @property
    def vertex_refs( self ):
        """ get list of cut indices for each vertex (stored in a numpy array) in the current base """
        return self._cell.vertex_refs( False )

    @property
    def nb_active_cuts( self ):
        """ beware: if remove_inactive_cuts() has not been called, this procedure will have first to compute the active cuts (may take some time) """
        return self._cell.nb_active_cuts()

    @property
    def nb_stored_cuts( self ):
        """ may be different from nb_active_cuts if remove_inactive_cuts() has not been called """
        return self._cell.nb_stored_cuts()

    @property
    def nb_vertices_td( self ):
        """ nb vertices in current "true" dimensionality """
        return self._cell.nb_vertices( True )

    @property
    def nb_vertices( self ):
        return self._cell.nb_vertices( False )

    @property
    def bounded( self ):
        return self._cell.bounded()

    @property
    def empty( self ):
        return self._cell.empty()
    
    # @property
    # def closed_faces( self ):
    #     return self._cell.closed_faces()
    
    @property
    def base( self ):
        """ internal base, used to get coordinates and directions if true_dimensionnality < nb_dims.
            It true_dimensionality == nb_dims, it return an identity matrix
        """
        return self._cell.base()
    
    @property
    def ndim( self ):
        return self._cell.ndim()
    
    @property
    def cuts( self ):
        return self._cell.cuts()
    
    def cut_boundary( self, dir, val, index = 0 ):
        return self._cell.cut_boundary( np.ascontiguousarray( dir ), val, int( index ) )

    def cut( self, dir, val, index = 0 ):
        return self._cell.cut( np.ascontiguousarray( dir ), val, int( index ) )

    def display_vtk( self, vtk_output ):
        return self._cell.display_vtk( vtk_output )

    # def for_each_face( self, on_face ):
    #     """
    #         Call args:
    #             on_face( cut_refs, vertex_indices, ray_refs_list ),
    #     """
    #     return self._cell.for_each_face( on_face )

    def for_each_edge( self, on_edge ):
        """
            on_edge( edge_refs, vertex_indices )
                * edge_refs = [ num_cut_0, ... ]
                * vertex_indices => 
                   - size = 2 for a regular edge
                   - size = 1 for a ray

        """
        return self._cell.for_each_edge( on_edge, False )

    def for_each_edge_td( self, on_edge ):
        """
            on_edge( edge_refs, vertex_indices )
                * edge_refs = [ num_cut_0, ... ]
                * vertex_indices => 
                   - size = 2 for a regular edge
                   - size = 1 for a ray

        """
        return self._cell.for_each_edge( on_edge, True )

    def ray_dir( self, ray_refs, base_vertex ):
        return self._cell.ray_dir( ray_refs, base_vertex )

    def plot_pyplot( self, fig, color = 'black', linestyle = '-', linewidth = 2, low_dim_linewidth_coeff = 0.25, ray_length = 0.2, ray_color = None, ray_linestyle = '--', free_color = None, free_linestyle = ':' ):
        """
            ray_length is the length used to display infinite edges.
        """

        # check fig properties
        if isinstance( fig, ModuleType ):
            fig.axis( 'equal' )

        if self.ndim == 3:
            fig = fig.figure()
            fig = fig.add_subplot( projection = '3d' )

        #
        coords = self.vertex_coords_td @ self.base

        # find a complementary base
        cbase = list( np.eye( self.ndim ) )
        for r in range( self.ndim ):
            for c in self.base:
                cbase[ r ] -= np.dot( cbase[ r ], c ) * c
        cbase.sort( key = np.linalg.norm )
        cbase = cbase[ self.true_dimensionality : ]

        # helper to plot
        def pl( array, gtype ):
            array = np.array( array )
            xyz = [ array[ :, d ] for d in range( min( self.ndim, 3 ) ) ]
            kwa = { 'linewidth': linewidth }
            if self.true_dimensionality < self.ndim and gtype != 'free':
                kwa[ 'linewidth' ] *= low_dim_linewidth_coeff

            if gtype == 'free':
                return fig.plot( *xyz, **kwa, linestyle = free_linestyle, color = free_color or color )
            if gtype == 'ray':
                return fig.plot( *xyz, **kwa, linestyle = ray_linestyle, color = ray_color or color )
            if gtype == 'edge':
                return fig.plot( *xyz, **kwa, linestyle = linestyle, color = color )

        # for each vertex, plot free items, get the edges and the rays
        for c in coords:
            for cvec in cbase:
                pl( [ c - ray_length * cvec, c + ray_length * cvec ], 'free' )

        # plot edge
        def edge_func( refs, vertices ):
            if len( refs ) == 0:
                return

            # ray
            if len( vertices ) == 1:
                rd = self.ray_dir( refs, vertices[ 0 ] ) @ self.base
                rd /= np.linalg.norm( rd )
                bc = coords[ vertices[ 0 ] ]
                pl( [ bc, bc + ray_length * rd ], 'ray' )
                return
            
            # edge
            if len( vertices ) == 2:
                pl( [ coords[ vertices[ 0 ] ], coords[ vertices[ 1 ] ] ], 'edge' )
                return

            raise RuntimeError( "not an expected size for vertices" )

        self.for_each_edge_td( edge_func )

    def plot( self, fig = None, **kwargs ):
        """  
        """
        if fig is None:
            import matplotlib.pyplot as plt
            self.plot( plt )
            plt.show()
            return

        self.plot_pyplot( fig, **kwargs )

        # if self.true_dimensionality == 2:

        #     def disp_ray( ray_refs, vertex_index, rev ):
        #         print( "ray" )
        #         dir = self.ray_dir( ray_refs, vertex_index ) @ self.base.T
        #         b_c = coords[ vertex_index ]
        #         d_c = b_c + ray_length * dir / np.linalg.norm( dir )
        #         x = [ b_c[ 0 ], d_c[ 0 ] ]
        #         y = [ b_c[ 1 ], d_c[ 1 ] ]
        #         if rev:
        #             x.reverse()
        #             y.reverse()
        #         fig.plot( x, y, ray_linestyle or linestyle, color = ray_color or color )

        #     def on_face( edge_refs, vertex_indices, ray_refs ):
        #         print( "cut_refs" )

        #         if len( ray_refs ) >= 1:
        #             disp_ray( ray_refs[ 0 ], vertex_indices[ 0 ], 1 )

        #         if len( vertex_indices ) >= 2:
        #             x = [ coords[ n, 0 ] for n in vertex_indices ]
        #             y = [ coords[ n, 1 ] for n in vertex_indices ]
        #             if len( ray_refs ) == 0: # closed
        #                 x.append( x[ 0 ] )
        #                 y.append( y[ 0 ] )
        #             fig.plot( x, y, linestyle, color = color )

        #         if len( ray_refs ) >= 2:
        #             disp_ray( ray_refs[ 1 ], vertex_indices[ -1 ], 0 )

        #     self.for_each_face( on_face )
        #     return

        # raise RuntimeError( f"TODO: plot for dim == { self.ndim }" )
        



from .Distribution import Distribution
from ..PoomVec import PoomVec

class SumOfDiracs( Distribution ):
    """
        sum of potentially weighted diracs


    """
    
    def __init__( self, positions, masses = None ) -> None:
        """
            By default, masses are equal to 1 / len( positions )
        """

        # init values
        self._positions = None

        # ctor arguments
        self.positions = positions

    @property
    def positions( self ):
        return self._positions

    @positions.setter
    def positions( self, values ):
        if values is None:
            return
        self._positions = PoomVec( values )
        
    @property
    def ndim( self ):
        return self._positions.shape[ 1 ]



from .bindings.loader import module_for
# import numpy as np

class VtkOutput:
    """ A simple class to generate .vtk files 
    
    """

    def __init__( self ):
        """ 
        """

        self._module = module_for( 'generic_objects' )
        self._inst = self._module.VtkOutput()


    def save( self, filename ):
        self._inst.save( filename )



from .bindings.loader import module_for, normalized_dtype
import numpy as np

class PoomVec:
    """ Potentially Out of Memory Vector.

        Basically a Wrapper around cpp sdot::PoomVec class 
    
    """

    def __init__( self, values = None, shape = None, dtype = None, _vec = None, _module = None ):
        """ 
        """

        # already a PoomVec ?
        if isinstance( values, PoomVec ):
            self._module = values._module
            self._vec = values._vec
            return
        
        # try to convert values to a numpy array
        try:
            if len( values ):
                array = np.ascontiguousarray( values )
                if array.ndim:
                    if shape is None:
                        shape = array.shape
                    if dtype is None:
                        dtype = array.dtype
                    scalar_type = normalized_dtype( dtype )

                    sizes = {}
                    for n in range( 1, len( shape ) ):
                        sizes[ f'size_{ n - 1 }'] = shape[ n ] 

                    self._module = module_for( f'poom_vec_nd_{ len( shape ) - 1 }', scalar_type = scalar_type, **sizes )

                    self._vec = self._module.make_PoomVec_from_ndarray( array )
                    return
        except TypeError as e:
            pass

        # specification of attriutes
        if _vec is not None:
            self._module = _module
            self._vec = _vec
            return
        
        # default values
        if values is None:
            self._module = None
            self._vec = None
            return
        
        # :P
        raise RuntimeError( f"don't known how to construct a PoomVec using the given arguments" )

    def __repr__( self ):
        return self._vec.__repr__()

    @property
    def as_ndarray( self ):
        """  """
        return self._vec.as_ndarray()

    @property
    def dtype( self ):
        """  """
        return self._vec.dtype()

    @property
    def shape( self ):
        """  """
        return self._vec.shape()

    @property
    def size( self ):
        """  """
        return self.shape[ 0 ]

    def __iadd__( self, other ):
        self._vec.self_add( PoomVec( other )._vec )
        return self

    def __isub__( self, other ):
        self._vec.self_sub( PoomVec( other )._vec )
        return self

    def __idiv__( self, other ):
        self._vec.self_div( other )
        return self




from .Distribution import Distribution
from ..Expr import Expr

class SymbolicFunction( Distribution ):
    """ 
    """

    def __init__( self, expr ):
        """ 
        """
        self.expr = Expr( expr )

    def __repr__( self ):
        return self.expr.__repr__()

    def boundary_split( self, ndim = None ):
        return self.expr.boundary_split( ndim )



from .SdotSolver import SdotSolver

def optimal_transport_plan( source_distribution, target_distribution, norm = 2, mass_error_ratio = 1e-4 ):
    """ 
        optimal transport plan between source_distribution and target_distribution
    """
    sp = SdotSolver( source_distribution, target_distribution, norm )
    sp.mass_error_ratio = mass_error_ratio
    sp.adjust_potentials()
    return sp



from .PowerDiagram import PowerDiagram, rt_int
import numpy as np
import vfs

class ConvexFunction:
    """ 
        Function =
          max( 
            -inf, 
            scalar_product( m_dirs[ 0 ], y ) - m_offs[ 0 ], 
            scalar_product( m_dirs[ 1 ], y ) - m_offs[ 1 ], 
            ...
          ) 
          for y such that 
            scalar_product( b_dirs[ 0 ], y ) - b_offs[ 0 ] <= 0,
            scalar_product( b_dirs[ 1 ], y ) - b_offs[ 1 ] <= 0,
            ...
    """

    def __init__( self, m_dirs, m_offs, b_dirs = None, b_offs = None, dtype = np.double ) -> None:
        # m_...
        self.m_dirs = np.asarray( m_dirs, dtype = dtype )
        self.m_offs = np.asarray( m_offs, dtype = dtype )

        assert( self.m_dirs.ndim == 2 )
        assert( self.m_offs.ndim == 1 )
        assert( self.m_offs.shape[ 0 ] == self.m_dirs.shape[ 0 ] )

        # b_...
        if b_dirs is None or not len( b_dirs ):
            assert( b_offs is None or not len( b_offs ) )
            b_dirs = np.zeros( [ 0, self.m_dirs.shape[ 1 ] ], dtype = dtype )
            b_offs = np.zeros( [ 0 ], dtype = dtype )
        self.b_dirs = np.asarray( b_dirs, dtype = dtype )
        self.b_offs = np.asarray( b_offs, dtype = dtype )

        assert( self.b_dirs.ndim == 2 )
        assert( self.b_offs.ndim == 1 )
        assert( self.b_offs.shape[ 0 ] == self.b_dirs.shape[ 0 ] )
        assert( self.b_dirs.shape[ 1 ] == self.m_dirs.shape[ 1 ] )

        # attributes by deduction
        self.nb_dims = self.m_dirs.shape[ 1 ]

    def make_approx_from_values_and_derivatives( sample_coords, f_val, f_der, b_dirs = None, b_offs = None ):
        m_dirs = []
        m_offs = []
        for num_point in range( sample_coords.shape[ 1 ] ):
            point = sample_coords[ :, num_point ]
            val = f_val( point )
            der = f_der( point )

            m_dirs.append( der )
            m_offs.append( np.dot( der, point ) - val )
        return ConvexFunction( m_dirs, m_offs, b_dirs, b_offs )

    def to_power_diagram( self ):
        weights = np.sum( self.m_dirs * self.m_dirs, axis = 1 ) - 2 * self.m_offs
        return PowerDiagram( self.m_dirs, weights, self.b_dirs, self.b_offs )

    def write_vtk( self, filename, fit_boundary = 1.0 ):
        pd = self.to_power_diagram()
        pd.write_vtk( filename, fit_boundary )

    def summary( self ):
        from sympy import Symbol, tensorproduct

        y = [ Symbol( f'y_{ d }' ) for d in range( self.nb_dims ) ]

        repr_max = lambda i: str( sum( tensorproduct( self.m_dirs[ i, : ], y ) ) -  self.m_offs[ i ] )
        repr_bnd = lambda i: str( sum( tensorproduct( self.b_dirs[ i, : ], y ) ) <= self.b_offs[ i ] )

        lst_max = [ repr_max( i ) for i in range( self.m_dirs.shape[ 0 ] ) ]
        lst_bnd = [ repr_bnd( i ) for i in range( self.b_dirs.shape[ 0 ] ) ]
        lst_max.sort()
        lst_bnd.sort()

        res = f"max({ ', '.join( lst_max ) })"
        if len( lst_bnd ):
            res += f" for { ', '.join( lst_bnd ) }"
        return res

    def legendre_transform( self ):
        # call the legendre_transform func
        func = vfs.function( 'legendre_transform', [ f'inc_file:sdot/legendre_transform/legendre_transform.h' ] )
        new_m_dirs, new_m_offs, new_b_dirs, new_b_offs = func( self.m_dirs, self.m_offs, self.b_dirs, self.b_offs, rt_int( self.nb_dims ) )
        new_m_dirs = np.reshape( new_m_dirs, [ -1, self.nb_dims ] )
        new_b_dirs = np.reshape( new_b_dirs, [ -1, self.nb_dims ] )

        # make a new ConvexFunction
        return ConvexFunction( new_m_dirs, new_m_offs, new_b_dirs, new_b_offs )

    def __repr__( self ):
        return self.summary()

    def is_inside_boundaries( self, x_k ):
        if isinstance( x_k, ( int, float ) ):
            x_k = [ x_k ]
        x_k = np.asarray( x_k )

        return self.b_dirs.shape[ 0 ] == 0 or ( self.b_dirs @ x_k - self.b_offs ).max() <= 0

    def __call__( self, x_k ):
        return self.val( x_k )
    
    def val( self, x_k ):
        if isinstance( x_k, ( int, float ) ):
            x_k = [ x_k ]
        x_k = np.asarray( x_k )

        if not self.is_inside_boundaries( x_k ):
            return float('inf')
        
        return ( self.m_dirs @ x_k - self.m_offs ).max()

    def grad( self, x_k ):
        if isinstance( x_k, ( int, float ) ):
            x_k = [ x_k ]
        x_k = np.asarray( x_k )

        if not self.is_inside_boundaries( x_k ):
            return float('inf')

        k = ( self.m_dirs @ x_k - self.m_offs ).argmax()     
        return self.m_dirs[ k ]

    def __add__( self, u2 ):
        u1 = self
        if isinstance( u2, ConvexFunction ): # addition of two convex functions
            if u1.nb_dims == u2.nb_dims:
                nb_dims = u1.nb_dims
            else:
                print('Dimensions do not match.')

            new_m_dirs = ( u1.m_dirs[ :, None, : ] + u2.m_dirs[ None, :, : ] ).reshape( ( -1, nb_dims ) )
            new_m_offs = ( u1.m_dirs[ :, None ] + u2.m_dirs[ None, : ] ).flatten()
            new_b_dirs = np.block( [ [ u1.b_dirs ], [ u2.b_dirs ] ] )
            new_b_offs = np.concatenate( [ u1.b_offs, u2.b_offs ] )

            return ConvexFunction( new_m_dirs, new_m_offs, new_b_dirs, new_b_offs )
            
        elif isinstance( u2, ( int, float ) ): # addition of a polyhedral function and a scalar
            return ConvexFunction( u1.m_dirs, u1.m_offs - u2, u1.b_dirs, u1.b_offs )
        
        else:
            raise NotImplementedError( "Can't add Polyhedral and {}".format( type( u2 ) ) )
        
    def __radd__( self, scal ):
        return self.__add__( scal )
    
    def __sub__( self, scal ):
        if isinstance( scal,( int, float ) ): 
            return ConvexFunction( self.m_dirs, self.m_offs + scal, self.b_dirs, self.b_offs )
        else:
            raise NotImplementedError( "Can't substract ConvexFunction from {}".format( type( scal ) ) )
    
    def __mul__( self, scal ):
        if isinstance( scal, ( int, float ) ): 
            assert( scal >= 0 )
            return ConvexFunction( scal * self.m_dirs, scal * self.m_offs, self.b_dirs, self.b_offs )
        else:
            raise NotImplementedError( "Can't multiply ConvexFunction and {}".format( type( scal ) ) )
            
    def __rmul__( self, scal ):
        return self.__mul__( scal )
        
    def __truediv__( self, scal ):
        if isinstance( scal, ( int, float ) ): 
            assert( scal > 0 )
            return ConvexFunction( self.m_dirs / scal, self.m_offs / scal, self.b_dirs, self.b_offs )
        else:
            raise NotImplementedError( "Can't divide {} by ConvexFunction.".format( type( scal ) ) )


    def __or__(self,u2): 
        u1 = self
        if isinstance( u2, ConvexFunction ): # maximum of two PHC functions
            new_m_dirs = np.block( [ [ u1.m_dirs ], [ u2.m_dirs ] ] )
            new_m_offs = np.concatenate( [ u1.m_offs, u2.m_offs ] )
            new_b_dirs = np.block( [ [ u1.b_dirs ], [ u2.b_dirs ] ] )
            new_b_offs = np.concatenate( [ u1.b_offs,u2.b_offs ] )
            return ConvexFunction( new_m_dirs, new_m_offs, new_b_dirs, new_b_offs )
            
        elif isinstance(u2,(int,float)): # addition of a polyhedral function and a scalar
            return u1.__or__( ConvexFunction( np.zero( [ 1, u1.nb_dims ] ), [ -u2 ] ) )
        
        else:
            raise NotImplementedError( "Can't take the maximum of ConvexFunction and {}".format( type( u2 ) ) )
        
    def __ror__( self, scal ):
        return self.__or__( scal )


    def __xor__( self, u2 ): # inf-convolution of two PHC functions
        if isinstance( u2, ConvexFunction ):
            return ( self.legendre_transform() + u2.legendre_transform() ).legendre_transform()
            
        else:
            raise NotImplementedError("Can't take the inf-convolution of ConvexFunction and {}".format(type(u2)))



from .loader import global_build_directory, module_for
from ..Expr import Expr
import math

def symbolic_integration( func, nb_dims ):
    if func == Expr( "1" ):
        return Expr( "loc_vol" )
    # / { math.factorial( nb_dims ) }
    #   simplex = []
    yo()
    return 1 / math.factorial( nb_dims )

def integration_module( funcs, scalar_type, nb_dims ):
    # get summary
    ct_repr, rt_data = Expr.ct_rt_split_of_list( funcs )
    module_name = 'integration_' + ct_repr

    # generate .cpp and SConstruct files
    bd = get_build_dir( 'generated', ct_repr )
    cf = bd / ( module_name + ".cpp" )
    with open( cf, 'w' ) as f:
        f.write( '#include <sdot/support/binding_config.h>\n' )
        f.write( '#include <sdot/Cell.h>\n' )
        
        f.write( '#include <pybind11/pybind11.h>\n' )
        f.write( '#include <pybind11/numpy.h>\n' )
        f.write( '#include <pybind11/stl.h>\n' )

        f.write( 'using namespace sdot;\n' )

        f.write( 'using Array_TF = pybind11::array_t<SDOT_CONFIG_scalar_type, pybind11::array::c_style>;\n' )
        f.write( 'using Array_PI = pybind11::array_t<PI, pybind11::array::c_style>;\n' )
        f.write( 'static constexpr int nb_dims = SDOT_CONFIG_nb_dims;\n' )
        f.write( 'using Arch = sdot::SDOT_CONFIG_arch;\n' )
        f.write( 'using TF = SDOT_CONFIG_scalar_type;\n' )
        f.write( 'using Pt = Vec<TF, nb_dims>;\n' )

        f.write( 'struct PD_NAME( CutInfo ) {\n' )
        f.write( '    CutType type;\n' )
        f.write( '    Pt      p1;\n' )
        f.write( '    TF      w1;\n' )
        f.write( '    PI      i1;\n' )
        f.write( '};\n' )
        f.write( '\n' )
        f.write( 'struct PD_NAME( CellInfo ) {\n' )
        f.write( '    Pt p0;\n' )
        f.write( '    TF w0;\n' )
        f.write( '    PI i0;\n' )
        f.write( '};\n' )

        f.write( 'using TCell = Cell<Arch,TF,nb_dims,PD_NAME( CutInfo ),PD_NAME( CellInfo )>;\n' )
        f.write( 'using TCut = Cut<TF,nb_dims,PD_NAME( CutInfo )>;\n' )

        splits, final_funcs = Expr.cell_splits_of_list( funcs, rt_data )

        f.write( 'struct PD_NAME( Integration ) {\n' )
        f.write( '    void operator()( const Vec<Pt,nb_dims+1> &simplex ) {\n' )
        f.write( '        using namespace std;\n' )
        f.write( '        Eigen::Matrix<TF,nb_dims,nb_dims> M;\n' )
        f.write( '        for( PI r = 0; r < nb_dims; ++r )\n' )
        f.write( '            for( PI c = 0; c < nb_dims; ++c )\n' )
        f.write( '                M( r, c ) = simplex[ r + 1 ][ c ] - simplex[ 0 ][ c ];\n' )
        f.write( f'        TF loc_vol = M.determinant() / { math.factorial( nb_dims ) };\n' )
        f.write( f'        TF coeff = loc_vol >= 0 ? 1 : -1;\n' )
        f.write( f'        vol += coeff * loc_vol;\n' )
        for i in range( len( final_funcs ) ):
           f.write( f'        out[ { i } ] += coeff * ( { symbolic_integration( final_funcs[ i ], nb_dims ) } );\n' )
        f.write( '    }\n' )
        f.write( f'    Vec<TF,{ len( final_funcs ) }> out;\n' )
        f.write( f'    TF vol;\n' )
        f.write( '};\n' )

        f.write( 'PYBIND11_MODULE( SDOT_CONFIG_module_name, m ) {\n' )
        f.write( '    m.def( "cell_integral", []( TCell &cell, const std::vector<ExprData> &rt_data ) -> std::vector<TF> {\n' )
        f.write( '        PD_NAME( Integration ) res;\n' )
        f.write( '        for( auto &v : res.out )\n' )
        f.write( '            v = 0;\n' )
        f.write( '        res.vol = 0;\n' )
        for split in splits:
            f.write( split[ 0 ] )
        f.write( '        cell.simplex_split( res );\n' )
        for split in reversed( splits ):
            f.write( split[ 1 ] )
        f.write( '        return { res.out.begin(), res.out.end() };\n' )
        f.write( '    } );\n' )

        # m.def( "measures", []( AccelerationStructure<TCell> &as, const TCell &base_cell, ConstantValue<TF> cv ) {
        #     pybind11::array_t<TF, pybind11::array::c_style> res( Vec<PI,1>{ as.nb_cells() } );
        #     as.for_each_cell( base_cell, [&]( TCell &cell, int num_thread ) {
        #         res.mutable_at( cell.info.i0 ) = cell.measure( cv );
        #     } );
        #     return res;
        # } );

        f.write( '}\n' )

    sf = bd / ( module_name + ".SConstruct" )
    with open( sf, 'w' ) as f:
        f.write( 'from sdot.bindings.construct import construct\n' )
        f.write( 'construct( Environment, VariantDir, Configure, ARGLIST, "' + module_name + '", [ "scalar_type", "nb_dims" ], [\n' )
        f.write( '    "' + str( cf ) + '",\n' )
        f.write( '] )\n' )

    module = module_for( module_name, dir_of_the_SConstruct_file = bd, scalar_type = scalar_type, nb_dims = nb_dims )
    return module, rt_data



# from .bindings.loader import sdot_module_for
from ..TransformationMatrix import TransformationMatrix
from .Distribution import Distribution
import numpy as np

class UnitBox( Distribution ):
    """ The mass (integral) of this box is prescribed and equal to 1 by default (hence the "Unit" in the name), even if the box is scaled.

        The default it is a [0,1]^ndim box, but one can apply an affine transformation to move it to another places.
    """

    def __init__( self, transformation = None, ndim = None, mass = 1 ):
        """ 
            `transformation` is a way to move the [0,1]^ndim basic box. It must be compatible with TransformationMatrix.
            `ndim` si optional (it can be useful if there aren't already other ways to find it in calling procedures)
            `mass` is the prescribed total mass (after the transformation)
        """

        self.transformation = TransformationMatrix( transformation )
        self._ndim = ndim
        self.mass = mass

    def boundary_split( self, ndim = None ):
        if ndim is None:
            if self.ndim is None:
                raise RuntimeError( 'found no way to guess the dimensionality' )
            ndim = self.ndim

        # TODO: include transformations
        bnds = []
        for d in range( ndim ):
            bnds.append( [ + ( i == d ) for i in range( ndim ) ] + [ 1 ] )
            bnds.append( [ - ( i == d ) for i in range( ndim ) ] + [ 0 ] )

        return bnds, 1
    
    @property
    def ndim( self ):
        return self._ndim


